#include<vcg/complex/complex.h>
#include<vcg/complex/algorithms/create/platonic.h>
#include<vcg/space/line2.h>
#include<vcg/space/intersection2.h>
#include<wrap/io_trimesh/import_off.h> 
#include<wrap/io_trimesh/export_off.h> 
#include<wrap/io_trimesh/export_ply.h> 

using namespace vcg;

class MyVertex; class MyEdge; class MyFace;
struct MyUsedTypes : public UsedTypes<Use<MyVertex>   ::AsVertexType,
                                           Use<MyEdge>     ::AsEdgeType,
                                           Use<MyFace>     ::AsFaceType>{};

class MyVertex  : public Vertex< MyUsedTypes, vertex::Coord3f, vertex::Normal3f, vertex::BitFlags  >{};
class MyFace    : public Face<   MyUsedTypes, face::FFAdj, face::Normal3f,  face::VertexRef, face::BitFlags > {};
class MyEdge    : public Edge<   MyUsedTypes> {};

class MyMesh    : public tri::TriMesh< std::vector<MyVertex>, 
    std::vector<MyFace> , std::vector<MyEdge>  > {};


const float EPS = 0.001f;

// Small class to represent a point on a triangle. 
struct PointOnTri
{
  PointOnTri(Point3f _p,int _e, MyFace *_f) : p(_p),edgeInd(_e),f(_f){}
  PointOnTri(Point3f _p, MyFace *_f) : p(_p),edgeInd(-1),f(_f){}
  Point3f p;
  Point3f n;  
  int edgeInd;
  MyFace *f;
  
  bool isValid();
};

bool PointOnTri::isValid()
{
  if(f==NULL) 
    return false;
  
  Point3f bc; // barycentric coordinates of input point
  bool ret = InterpolationParameters(*f, p, bc);
  if(!ret)
    return false;
  if(!(bc[0]<1.01 && bc[1]<1.01 && bc[2]<1.01)) 
    return false; // check that the input point is inside the tri.
  
  // if the point is on an edge, baricentric coords of the vertex on the other side should be 0;
  // eg if the point is on egde 0 (v0-v1) then is bc[2] ~ 0
  // eg if the point is on egde 1 (v1-v2) then is bc[0] ~ 0
  // eg if the point is on egde 2 (v2-v0) then is bc[1] ~ 0
  if(edgeInd >=0)
  {
    if(fabs(bc[(edgeInd+2)%3]) > EPS)
      return false;
  }
  return true;
}

// Given the normal of a triangle and the direction of gravity
// returns the projected gravity direction onto the triangle plane
Point3f TangentGravity(const Point3f n, const Point3f g)
{
  Point3f l = n ^ g;
  float len = Norm(l); 
  if(len<EPS) printf("warning...\n");
  l.Normalize();
  Point3f t = l ^ n;      
  t.Normalize();
  return t;  
}


Point3f RandomTri(const Point3f &p0, const Point3f &p1, const Point3f &p2)
{
 float u = float(rand())/float(RAND_MAX);
 float v = float(rand())/float(RAND_MAX);
 if(u+v>1.f) {u=1.f-u; v=1.f-v;}
 
 return p0+(p1-p0)*u+(p2-p0)*v;
}

// Fill a vector with <vertNum> particles uniformly distributed over the surface. 
void initPointVec(MyMesh &m, int vertNum, std::vector< PointOnTri > &pointVec, Point3f &g)
{
  pointVec.clear();
  while(pointVec.size()<vertNum)
  {
    MyFace &f=m.face[rand()%m.FN()];
    Point3f p = RandomTri(f.P(0),f.P(1),f.P(2));  
    pointVec.push_back(PointOnTri(p,&f));
    pointVec.back().n = TangentGravity(f.N(),g);
    assert(pointVec.back().isValid());
  }
}
 
bool movePointToEdge(PointOnTri &pot, const Point3f &gravity) 
{
  MyFace *f = pot.f;
  Point3f p = pot.p;
  Point3f tg = TangentGravity(pot.f->N(),gravity);

  const int bestProjDir = vcg::Abs( pot.f->N() ).MaxCoeffId();
  const int xi = (bestProjDir+1)%3;
  const int yi = (bestProjDir+2)%3;
  
  Point2f vp2[3]; // 2D positions of the vertexes of the triangle projected on the XY plane
  for(int j=0;j<3;++j)
    vp2[j] = Point2f(f->P(j)[xi],f->P(j)[yi]);
  
  Point2f p2 = Point2f(p[xi],p[yi]);  // the projection of the input point in 2D 
  Point2f tg2 = Point2f(tg[xi],tg[yi]);  // the projection of the t vector onto the proj plane
  Line2f dir(p2,tg2);                // 2D line from point along the t direction 
  
  Point2f ip[3];   // intersections of the edge lines with the line dir
  int bestSide = -1;
  float minDist = std::numeric_limits<float>::max();
  for(int j=0;j<3;++j) if(j!=pot.edgeInd)
  {
    Line2f edgeLine = Line2f(vp2[j],vp2[(j+1)%3]-vp2[j]);
    bool res= LineLineIntersection(dir,edgeLine,ip[j]);
    if(res)
    {
        Point3f bci;
        bool ret = InterpolationParameters2(vp2[0],vp2[1],vp2[2], ip[j], bci);
        if(bci[0]>-EPS && bci[1]>-EPS && bci[2]>-EPS)
          if(Distance(ip[j],p2) < minDist)
          {
            minDist = Distance(ip[j],p2);
            bestSide=j;            
          }            
    }
  }
  assert(bestSide != -1);
    
  // riportare in 3D l'intersezione 2D tra la linea e gli edge.     
  Point3f bci;
  bool ret = InterpolationParameters2(vp2[0],vp2[1],vp2[2], ip[bestSide], bci);
  Point3f newp = f->P(0)*bci[0] + f->P(1)*bci[1] + f->P(2)*bci[2];
  
  pot.edgeInd = pot.f->FFi(bestSide);
  pot.f = pot.f->FFp(bestSide);
  pot.p=newp;  
  assert(pot.isValid());
  
  return bestSide;
}
int main( int /*argc*/, char **argv )
{
  MyMesh m;
  MyMesh out;
  Point3f myGravity(1,0,0);
  std::vector<PointOnTri > pointVec;
  
  tri::io::ImporterOFF<MyMesh>::Open(m,argv[1]);
  tri::UpdateNormal<MyMesh>::PerVertexNormalizedPerFaceNormalized(m);
  tri::UpdateTopology<MyMesh>::FaceFace(m);
  
  printf("Input mesh  vn:%i fn:%i\n",m.VN(),m.FN());
  
  initPointVec(m,50,pointVec,myGravity);
  
  for(auto pr=pointVec.begin();pr!=pointVec.end();++pr)
    tri::Allocator<MyMesh>::AddVertex(out,pr->p); 
  
  for(int i=0;i<15;++i)
  {
    for(auto pr=pointVec.begin();pr!=pointVec.end();++pr)
      movePointToEdge(*pr,myGravity);
    
    for(auto pr=pointVec.begin();pr!=pointVec.end();++pr)
      tri::Allocator<MyMesh>::AddVertex(out,pr->p); 
  }
  
  tri::io::ExporterPLY<MyMesh>::Save(out,"out.ply",tri::io::Mask::IOM_VERTNORMAL);
  return 0;
}
